@@ -6,29 +6,26 @@
 """
 Provides implementation of torch.utils.data.Dataset for word-level input.
 """
+import logging
+import sys
+#logging.basicConfig(level=logging.ERROR)
+#if not sys.warnoptions:
+#    import warnings
+#    warnings.simplefilter('ignore')
+#    os.environ['PYTHONWARNINGS'] = 'ignore'
 from junky import CPU, absmax_torch, pad_array_torch, \
                   pad_sequences_with_tensor
 from junky.dataset.base_dataset import BaseDataset
-import sys
 from time import time
 import torch
 from torch import Tensor, tensor
 from tqdm import tqdm
 
+# to suppress transformers' warnings
+#logging.getLogger('transformers.tokenization_utils_base').setLevel(logging.ERROR)
+#logging.getLogger('pytorch_pretrained_bert.tokenization').setLevel(logging.ERROR)
+logging.getLogger('transformers').setLevel(logging.ERROR)
 
-# from collections import Iterable
-#
-# def print_struct(o, shift=0):
-#     if isinstance(o, torch.Tensor):
-#         print('{}tensor({}, dtype={})'
-#                   .format(' ' * shift, o.shape, o.dtype))
-#     elif isinstance(o, Iterable):
-#         print(' ' * (shift - 1), type(o), 'len =', len(o))
-#         for o_ in o:
-#             print_struct(o_, shift=shift + 4)
-#
-# def print_list(lst):
-#     [print('   ', x) for x in lst]
 
 class BertDataset(BaseDataset):
     """
@@ -46,7 +43,7 @@ class BertDataset(BaseDataset):
         sentences: sequences of words: list([list([str])]). If not ``None``,
             they will be transformed and saved. NB: All the sentences must
             not be empty.
-        All other args are params for the `.transpose()` method. They are used
+        All other args are params for the `.transform()` method. They are used
             only if *sentences* is not ``None``. You can use any args but
             `save` that is set to `True`.
     """
@@ -57,11 +54,22 @@ class BertDataset(BaseDataset):
 
     @property
     def vec_size(self):
-        return self.data[0].shape[-1] if self.data else \
-               self.model.config.hidden_size
+        if self.data:
+            if isinstance(self.data[0], list):
+                res = self.data[0][0].shape[-1]
+            else:
+                res = self.data[0].shape[-1]
+        else:
+            res = self.model.config.hidden_size
+        return res
 
     def __init__(self, model, tokenizer, int_tensor_dtype=torch.int64,
                  sentences=None, **kwargs):
+        assert model.config.output_hidden_states, \
+               'ERROR: BERT model was created with ' \
+               '`output_hidden_states=False` option. To use it in ' \
+               'BertDataset, the model must be created with ' \
+               '`output_hidden_states=True`'
         super().__init__()
         self.model = model
         self.tokenizer = tokenizer
@@ -89,14 +97,16 @@ class BertDataset(BaseDataset):
             else:
                 hiddens = [hidden_states[x] for x in layer_ids]
 
-            if aggregate_op == 'cat':
+            if aggregate_op == 'absmax':
+                hiddens = absmax_torch(hiddens, dim=0)
+            elif aggregate_op == 'cat':
                 hiddens = torch.cat(hiddens, dim=-1)
             elif aggregate_op == 'first':
                 hiddens = hiddens[0]
             elif aggregate_op == 'last':
                 hiddens = hiddens[-1]
             elif aggregate_op == 'max':
-                hiddens = absmax_torch(hiddens, dim=0)
+                hiddens = torch.max(hiddens, dim=0)[0]
             elif aggregate_op == 'mean':
                 hiddens = torch.mean(hiddens if isinstance(hiddens, Tensor) else
                                      torch.stack(hiddens, dim=0),
@@ -107,45 +117,47 @@ class BertDataset(BaseDataset):
                                      dim=0)
             else:
                 RuntimeError(
-                  'ERROR: unknown aggregate_op '
-                  "(choose one of ['cat', 'max', 'mean', 'sum'])"
+                    'ERROR: unknown aggregate_op '
+                    "(choose one of ['absmax', 'cat', 'max', 'mean', 'sum'])"
                 )
         return hiddens
 
     def transform(self, sentences, max_len=None, batch_size=64,
                   hidden_ids=0, aggregate_hiddens_op='mean',
-                  aggregate_subtokens_op='max', to=None,
-                  save=True, append=False, loglevel=1):
+                  aggregate_subtokens_op='absmax', to=None,
+                  with_grad=False, save=True, append=False, loglevel=1):
         """Convert *sentences* of words to the sequences of the corresponding
         contextual vectors and adjust their format for Dataset.
 
         *max_len* is a param for tokenizer. We'll transform lines of any
             length, but the quality is higher if *max_len* is greater.
             ``None`` (default) or `0` means the maximum for the model
-            (usually, 512).
+            (usually, `512`).
 
         *batch_size* affects only on the execution time. Greater is faster,
             but big *batch_size* may be cause of CUDA Memory Error. If
-            ``None`` or 0, we'll try to convert all *sentences* with one
+            ``None`` or `0`, we'll try to convert all *sentences* with one
             batch.
 
         *hidden_ids*: hidden score layers that we need to aggregate. Allowed
             int or tuple of ints. If ``None``, we'll aggregate all the layers.
 
         *aggregate_hidden_op*: how to aggregate hidden scores. The ops
-            allowed: 'cat', 'max', 'mean', 'sum'. For 'max' method we take
-            into account the absolute values of the compared items (absmax
-            method).
+            allowed: 'absmax', 'cat', 'max', 'mean', 'sum'. For the 'absmax'
+            method we take into account absolute values of the compared items.
 
         *aggregate_subtokens_op*: how to aggregate subtokens vectors to form
             only one vector for each input token. The ops allowed: ``None``,
-            'first', 'last', 'max', 'mean', 'sum'. For 'max' method we take
-            into account the absolute values of the compared items (absmax
-            method).
+            'absmax', 'append', 'first', 'last', 'max', 'mean', 'sum'. For the
+            'absmax' method we take into account absolute values of the
+            compared items.
 
         If you want to get the result placed on some exact device, specify the
         device with *to* param. If *to* is ``None`` (defautl), data will be
-        placed to the very device that `bs.model` is used.
+        placed to the very device that `ds.model` is used.
+
+        *with_grad*: calculate gradients during forward propagation through
+        self.model. Default is ``False``.
 
         If *save* is ``True``, we'll keep the converted sentences as the
         `Dataset` source.
@@ -162,16 +174,24 @@ class BertDataset(BaseDataset):
         *aggregate_subtokens_op* is used, each sentence will be converted to
         exactly one tensor of shape [<sentence length>, <vector size>]."""
 
+        tokenizer_max_len = self.tokenizer.max_len \
+                                if hasattr(self.tokenizer, 'max_len') else \
+                            self.tokenizer.model_max_len \
+                                if hasattr(self.tokenizer,
+                                           'model_max_len') else \
+                            512
+
         if not max_len:
-            max_len = self.tokenizer.max_len
+            max_len = tokenizer_max_len
         assert max_len >= 16, 'ERROR: max len must be >= 16'
-        assert max_len <= self.tokenizer.max_len, \
-               'ERROR: max len must be <= {}'.format(self.tokenizer.max_len)
-        valid_ops = ['cat', 'max', 'mean', 'sum']
+        assert max_len <= tokenizer_max_len, \
+               'ERROR: max len must be <= {}'.format(tokenizer_max_len)
+        valid_ops = ['absmax', 'cat', 'max', 'mean', 'sum']
         assert aggregate_hiddens_op in valid_ops, \
                'ERROR: unknown aggregate_hidden_op (choose one of {})' \
                    .format(valid_ops)
-        valid_ops = [None, 'first', 'last', 'max', 'mean', 'sum']
+        valid_ops = [None, 'absmax', 'append', 'first', 'last', 'max', 'mean',
+                     'sum']
         assert aggregate_subtokens_op in valid_ops, \
                'ERROR: unknown aggregate_subtokens_op (choose one of {})' \
                    .format(valid_ops)
@@ -197,12 +217,15 @@ class BertDataset(BaseDataset):
         _src = sentences
         if loglevel >= 2:
             print('Tokenizing')
-            _src = tqdm(_src, file=sys.stdout)
+            _src = tqdm(iterable=_src, mininterval=2, file=sys.stdout)
         time0 = time()
         # tokenize each token separately by the BERT tokenizer
-        tokenized_sentences = [[self.tokenizer.tokenize(x) or ['[UNK]']
-                                    for x in x]
-                                   for x in _src]
+        tokenized_sentences = \
+            [[x for x in self.tokenizer.tokenize(x) or ['[UNK]']]
+                 for x in _src] \
+                if sentences and isinstance(sentences[0], str) else \
+            [[self.tokenizer.tokenize(x) or ['[UNK]'] for x in x] \
+                 for x in _src]
         if time() - time0 < 2 and loglevel == 1:
             loglevel = 0
 
@@ -229,9 +252,12 @@ class BertDataset(BaseDataset):
         # lengths of flattened tokenized_sentences
         sent_lens = [len(x) for x in tokenized_sentences]
 
+        need_sort_dataset = self.sort_dataset \
+                        and self.use_batch_max_len \
+                        and len(tokenized_sentences) > batch_size
         ## sort tokenized sentences by lenght
         ######
-        if self.sort_dataset:
+        if need_sort_dataset:
             sent_lens, sorted_sent_ids = zip(*sorted(
                 [(x, i) for i, x in enumerate(sent_lens)],
                 reverse=True
@@ -311,25 +337,6 @@ class BertDataset(BaseDataset):
 
         splitted_sent_lens = [len(x) for x in tokenized_sentences]
 
-#         a = []
-#         print()
-#         print('overmap', overmap)
-#         print()
-#         for i, sent in enumerate(tokenized_sentences):
-#             if i in overmap:
-#                 j, pos = overmap[i]
-#                 start = token_starts[j][pos]
-#                 print('j', j, 'pos', pos, 'start', start)
-#                 print(a[j], len(a[j]))
-#                 print('   ', sent, len(sent))
-#                 a[j][start:] = sent
-#                 print('   ', a[j], len(a[j]))
-#                 print()
-#             else:
-#                 a.append(sent)
-#         print('\nRESTORE:')
-#         print_list(a)
-
         if not batch_size:
             batch_size = num_sents
 
@@ -337,9 +344,21 @@ class BertDataset(BaseDataset):
         _src = range(0, num_sents, batch_size)
         if loglevel:
             print('Vectorizing')
-            _src = tqdm(_src, file=sys.stdout)
-        for batch_i in _src:
+            _src = tqdm(iterable=_src, mininterval=2, file=sys.stdout)
 
+        ####### workaround for transformers v.4
+        test_enc = self.tokenizer.encode_plus(
+            text='', add_special_tokens=True, max_length=3,
+            padding='max_length', return_tensors=None,
+            return_token_type_ids=False, return_attention_mask=True
+        )
+        cls_id, sep_id, pad_id = test_enc['input_ids']
+        att_mask1, _, att_mask0 = test_enc['attention_mask']
+        #######
+
+        for batch_i in _src:
+            ####### workaround for transformers v.4
+            '''
             batch_max_len = min(
                 max(splitted_sent_lens[batch_i:batch_i + batch_size]) + 2,
                 max_len
@@ -349,24 +368,62 @@ class BertDataset(BaseDataset):
                 self.tokenizer.encode_plus(text=sent,
                                            add_special_tokens=True,
                                            max_length=batch_max_len,
-                                           pad_to_max_length=True,
+                                           #truncation=True,
+                                           #pad_to_max_length=True,
+                                           padding='longest',
                                            return_tensors='pt',
                                            return_attention_mask=True,
                                            return_overflowing_tokens=False)
                     for sent in tokenized_sentences[batch_i:batch_i
-                                                  + batch_size]
+                                                          + batch_size]
             ]
+            encoded_sentences = []
+            for sent in tokenized_sentences[batch_i:batch_i + batch_size]:
+                try:
+                    self.tokenizer.encode_plus(text=sent,
+                                               add_special_tokens=True,
+                                               max_length=batch_max_len,
+                                               #truncation=True,
+                                               #pad_to_max_length=True,
+                                               padding='longest',
+                                               return_tensors='pt',
+                                               return_attention_mask=True,
+                                               return_overflowing_tokens=False)
+                except TypeError as e:
+                    print('batch_max_len =', batch_max_len)
+                    print('type(sent) =', type(sent))
+                    print('sent = [{}]'.format(sent))
+                    raise e'''
+
+            batch_max_len = min(
+                max(splitted_sent_lens[batch_i:batch_i + batch_size]),
+                max_len - 2
+            ) if self.use_batch_max_len else max_len - 2
+
+            input_ids = tensor(
+                [([cls_id]
+                + self.tokenizer.convert_tokens_to_ids(sent)
+                + [sep_id] + [pad_id] * (batch_max_len - len(sent)))
+                     for sent in tokenized_sentences[batch_i:batch_i
+                                                           + batch_size]],
+                dtype=self.int_tensor_dtype
+            )
+            attention_mask = (input_ids != pad_id).type(self.int_tensor_dtype)
+            '''
             input_ids, attention_masks = zip(*[
                 (x['input_ids'], x['attention_mask'])
                     for x in encoded_sentences
-            ])
+            ])'''
+            #######
 
-            with torch.no_grad():
+            with torch.set_grad_enabled(with_grad):
                 hiddens = self.model(
-                    torch.cat(input_ids, dim=0).to(device),
+                    #torch.cat(input_ids, dim=0).to(device),
+                    input_ids.to(device),
                     token_type_ids=None,
-                    attention_mask=torch.cat(attention_masks, dim=0)
-                                        .to(device)
+                    #attention_mask=torch.cat(attention_masks, dim=0)
+                    #                    .to(device)
+                    attention_mask=attention_mask.to(device)
                 )[-1]
 
                 hiddens = self._aggregate_hidden_states(
@@ -418,7 +475,7 @@ class BertDataset(BaseDataset):
 
         ## sort data in original order
         ######
-        if self.sort_dataset:
+        if need_sort_dataset:
             _, data = zip(*sorted(
                 [(i, x) for i, x in enumerate(data)],
                 key=lambda x: sorted_sent_ids[x[0]]
@@ -426,28 +483,29 @@ class BertDataset(BaseDataset):
             data = list(data)
         ######
 
-        _src = num_subtokens
-        if loglevel:
-            print('Adjusting')
-            _src = tqdm(_src, file=sys.stdout)
-        for i, token_lens in enumerate(_src):
-            token_lens = num_subtokens[i]
-            start = 0
-            sent = data[i]
-            sent_ = []
-            for token_len in token_lens:
-                end = start + token_len
-                vecs = sent[start:end]
-                if aggregate_subtokens_op != None:
-                    vecs = self._aggregate_hidden_states(
-                        vecs, layer_ids=None,
-                        aggregate_op=aggregate_subtokens_op
-                    )
-                sent_.append(vecs)
-                start = end
+        if aggregate_subtokens_op != 'append':
+            _src = num_subtokens
+            if loglevel:
+                print('Reordering')
+                _src = tqdm(iterable=_src, mininterval=2, file=sys.stdout)
+            for i, token_lens in enumerate(_src):
+                #token_lens = num_subtokens[i]
+                start = 0
+                sent = data[i]
+                sent_ = []
+                for token_len in token_lens:
+                    end = start + token_len
+                    vecs = sent[start:end]
+                    if aggregate_subtokens_op != None:
+                        vecs = self._aggregate_hidden_states(
+                            vecs, layer_ids=None,
+                            aggregate_op=aggregate_subtokens_op
+                        )
+                    sent_.append(vecs)
+                    start = end
 
-            data[i] = sent_ if aggregate_subtokens_op == None else \
-                      torch.stack(sent_, dim=0)
+                data[i] = sent_ if aggregate_subtokens_op == None else \
+                          torch.stack(sent_, dim=0)
 
         if save:
             if append:
@@ -457,19 +515,23 @@ class BertDataset(BaseDataset):
         else:
             return data
 
-    def _collate(self, batch, with_lens=True, with_token_lens=True):
+    def _collate(self, batch, with_lens=True, with_token_lens=True,
+                 append_subtokens=False):
         """The method to use with `torch.utils.data.DataLoader` and
         `.transform_collate()`.
 
         :with_lens: return lengths of data.
-        :with_token_lens: return lengths of tokens of the data.
+        :with_token_lens: return lengths of tokens of the data (only allowed
+            if `.transform()` was called with `aggregate_subtokens_op=None`).
+        :append_subtokens: line up tokens if `.transform()` was called with
+            `aggregate_subtokens_op=None`. The result will be exactly as if
+            `aggregate_subtokens_op='append'` but you can get token lens.
         :return: depends on keyword args.
         :rtype: if the `.transform()` method was called with
             *aggregate_subtokens_op*=None:
                 tuple(list([torch.tensor]), lens:torch.tensor,
                       token_lens:list([torch.tensor]))
             otherwise: tuple(list([torch.tensor]), lens:torch.tensor)
-
         """
         device = CPU
         pad = 0.
@@ -488,13 +550,24 @@ class BertDataset(BaseDataset):
                         device = x[0].get_device()
                     tensor_dtype = x[0].dtype
                     break
-            lens = [tensor([len(x) for x in batch], device=device,
-                           dtype=self.int_tensor_dtype)] if with_lens else []
+            lens = [tensor([len([x for x in x for x in x]) for x in batch],
+                           device=device, dtype=self.int_tensor_dtype)] \
+                       if with_lens and append_subtokens else \
+                   [tensor([len(x) for x in batch], device=device,
+                           dtype=self.int_tensor_dtype)] \
+                       if with_lens else \
+                   []
             if with_token_lens:
                 lens.append([tensor([len(x) for x in x], device=device,
                                     dtype=self.int_tensor_dtype)
                                  for x in batch])
-            x = pad_array_torch(batch, padding_value=pad,
-                                device=device, dtype=tensor_dtype)
+            if append_subtokens:
+                x = pad_sequences_with_tensor(
+                    [torch.vstack([x for x in x for x in x]) for x in batch],
+                    padding_tensor=pad
+                )
+            else:
+                x = pad_array_torch(batch, padding_value=pad,
+                                    device=device, dtype=tensor_dtype)
 
         return (x, *lens)
