@@ -12,10 +12,12 @@ from gensim.models import keyedvectors
 import itertools
 import junky
 import numpy as np
-from sklearn.metrics import accuracy_score, f1_score, precision_score, \
-                            recall_score
+from sklearn.metrics import accuracy_score, f1_score, \
+                            precision_score, recall_score
 import sys
+import time
 import torch
+import torch.nn as nn
 from torch.utils.data import DataLoader, Dataset
 from tqdm import tqdm
 import warnings
@@ -33,6 +35,26 @@ def clear_stderr():
                   DeprecationWarning)
     clear_tqdm()
 
+
+def seconds_to_time(sec, with_week=True, round_sec=True):
+    s0 = int(round(sec)) if round_sec else sec
+    m0, s1 = divmod(int(s0), 60)
+    h0, m1 = divmod(m0, 60)
+    d0, h1 = divmod(h0, 24)
+    w, d1 = divmod(d0, 7) if with_week else (0, d0)
+    return (w, d1, h1, m1, s1), \
+           (w, d0, h0, m0, s0), \
+           (sec / 604800, sec / 86400, sec / 3600, sec / 60, sec)
+
+def seconds_to_strtime(sec, with_week=True, round_sec=True):
+    w, d, h, m, s = seconds_to_time(sec, with_week=with_week,
+                                    round_sec=round_sec)[0]
+    return '{}{}{}{}{}s'.format('{}w '.format(w) if w else '',
+                                '{}d '.format(d) if w + d else '',
+                                '{}h '.format(h) if w + d + h else '',
+                                '{}m '.format(m) if w + d + h + m else '',
+                                s)
+
 def make_word_embeddings(vocab, vectors=None,
                          pad_token=None, extra_tokens=None,
                          with_layer=True, layer_freeze=True, **layer_kwargs):
@@ -296,6 +318,8 @@ def extract_conllu_fields(corpus, fields=None, word2idx=None, unk_token=None,
     for i, sent in enumerate(corpus):
         if isinstance(sent, tuple):
             sent = sent[0]
+
+        isempty = True
         for j, field in enumerate(zip(*[
             (x['FORM'] if not word2idx or x['FORM'] in word2idx else
              unk_token,
@@ -309,12 +333,10 @@ def extract_conllu_fields(corpus, fields=None, word2idx=None, unk_token=None,
                                   and (not word2idx or x['FORM'] in word2idx
                                                     or unk_token)
         ])):
-            if field or with_empty:
-                sents[j].append(field)
-            else:
-                if return_nones:
-                    empties.append(i)
-                break
+            sents[j].append(field)
+            isempty = False
+        if isempty and return_nones:
+            empties.append(i)
 
         if return_nones:
             for j, x in enumerate(sent):
@@ -369,52 +391,65 @@ def to_device(data, device):
         data = data.to(device)
     elif isinstance(data, torch.nn.Module):
         data.to(device)
-    elif isinstance(data, Iterable):
+#     elif isinstance(data, Iterable):
+#         #data = type(data)(to_device(x, device) for x in data)
+    elif isinstance(data, (tuple, list)):
         data = type(data)(to_device(x, device) for x in data)
+    elif isinstance(data, tuple):
+        data = type(data)({x: to_device(y, device) for x, y in data.items()})
     return data
 
 def train(loaders, model, criterion, optimizer, scheduler,
           best_model_backup_method, log_prefix='', datasets=None,
           pad_collate=None, epochs=None, min_epochs=0, bad_epochs=5,
           batch_size=32, control_metric='accuracy', max_grad_norm=None,
-          best_score=None, with_progress=True, log_file=LOG_FILE):
+          batch_to_device=True, best_score=None, with_progress=True,
+          log_file=LOG_FILE):
 
     assert epochs or bad_epochs, \
-           'ERROR: Whether epochs or bad_epochs must be specified'
-    assert control_metric in ['accuracy', 'f1', 'loss'], \
-           "ERROR: unknown control_metric '{}' ".format(control_metric) \
-         + "(only 'accuracy', 'f1' and 'loss' are available)"
+           'ERROR: Either epochs or bad_epochs must be specified.'
+    assert control_metric in ['accuracy', 'f1', 'loss',
+                              'precision', 'recall'], \
+          f"ERROR: Unknown control_metric '{control_metric}' " \
+           "(only 'accuracy', 'f1', 'loss', 'precision' and 'recall' are " \
+           'available).'
     assert loaders or datasets, \
-           'ERROR: You must pass a tuple of Dataloader or Dataset ' \
-           'instances for train and test goals'
+           'ERROR: You must pass a list of Dataloader or Dataset ' \
+           'instances for train and test goals.'
 
     train_loader = loaders[0] if loaders and loaders[0] else \
                    datasets[0].create_loader(batch_size=batch_size,
                                              shuffle=True, num_workers=0) \
-                       if callable(getattr(datasets[0],
-                                           'create_loader', None)) else \
+                       if datasets and callable(getattr(datasets[0],
+                                                        'create_loader',
+                                                        None)) else \
                    DataLoader(datasets[0], batch_size=batch_size,
                               shuffle=True, num_workers=0,
                               collate_fn=pad_collate if pad_collate else
-                              datasets[0].pad_collate)
+                                         datasets[0].pad_collate) \
+                       if datasets and datasets[0] else None
     test_loader = loaders[1] \
                       if loaders and len(loaders) > 1 and loaders[1] else \
                   datasets[1].create_loader(batch_size=batch_size,
                                             shuffle=False, num_workers=0) \
                       if datasets and len(datasets) > 1 and datasets[1] \
-                                  and callable(
-                                      getattr(datasets[1],
-                                              'create_loader', None)
-                                  ) else \
+                                  and callable(getattr(datasets[1],
+                                                       'create_loader',
+                                                       None)) else \
                   DataLoader(datasets[1], batch_size=batch_size,
                              shuffle=False, num_workers=0,
                              collate_fn=pad_collate if pad_collate else
-                             datasets[1].pad_collate) \
+                                        datasets[1].pad_collate) \
                       if datasets and len(datasets) > 1 and datasets[1] else \
                   None
+    assert train_loader or test_loader, \
+           'ERROR: Either `loaders` or `dataset` must be not None.'
     assert test_loader or epochs, \
            'ERROR: At least one of the params `loaders[1]`, `dataset[1]` ' \
-           'or `epochs` must be not None'
+           'or `epochs` must be not None.'
+
+    is_bce = isinstance(criterion, (nn.BCEWithLogitsLoss, nn.BCELoss))
+    flatten_idx = -1 if is_bce else -2
 
     if not callable(best_model_backup_method):
         f = best_model_backup_method
@@ -439,55 +474,66 @@ def train(loaders, model, criterion, optimizer, scheduler,
     f1s = []
     bad_epochs_ = 0
     score = None
+    start_time = time.time()
 
     if with_progress:
         clear_tqdm()
     print_indent = ' ' * len(log_prefix)
     print_str = ''
     for epoch in range(1, epochs + 1) if epochs else itertools.count(start=1):
-        print_str = '{}Epoch {}: \n'.format(log_prefix, epoch)
-        train_losses_ = []
+        if train_loader:
+            print_str = '{}Epoch {}: \n'.format(log_prefix, epoch)
+            train_losses_ = []
 
-        progress_bar = tqdm(total=len(train_loader.dataset),
-                            desc='Epoch {}'.format(epoch),
-                            file=log_file) \
-                           if with_progress and log_file else \
-                       None
+            progress_bar = tqdm(desc='Epoch {}'.format(epoch),
+                                total=len(train_loader.dataset),
+                                file=log_file) \
+                               if with_progress and log_file else \
+                           None
 
-        model.train()
-        for batch in train_loader:
-            batch = to_device(batch, device)
-            optimizer.zero_grad()
-            pred, gold = model(*batch[:-1]), batch[-1]
+            model.train()
+            t, n_update = time.time(), 0
+            for batch in train_loader:
+                batch, gold = batch[:-1], to_device(batch[-1], device)
+                if batch_to_device:
+                    batch = to_device(batch, device)
+                optimizer.zero_grad()
+                pred = model(*batch)
 
-            batch_loss = []
-            for i in range(pred.size(0)):
-                tmp_loss = criterion(pred[i], gold[i])
-                batch_loss.append(tmp_loss)
+                if is_bce:
+                    gold = gold.float()
+                loss = criterion(pred.flatten(end_dim=flatten_idx),
+                                 gold.flatten(end_dim=-1))
+                loss.backward()
 
-            loss = torch.mean(torch.stack(batch_loss))
-            loss.backward()
+                if max_grad_norm:
+                    torch.nn.utils.clip_grad_norm_(
+                        parameters=model.parameters(), max_norm=max_grad_norm
+                    )
 
-            if max_grad_norm:
-                torch.nn.utils.clip_grad_norm_(parameters=model.parameters(),
-                                               max_norm=max_grad_norm)
+                optimizer.step()
+                train_losses_.append(loss.item())
 
-            optimizer.step()
-            train_losses_.append(loss.item())
+                if with_progress:
+                    t_ = time.time()
+                    n_update += len(gold)
+                    if t_ - t >= 2:
+                        t = t_
+                        progress_bar.set_postfix(
+                            train_loss=np.mean(train_losses_[-500:])
+                        )
+                        progress_bar.update(n_update)
+                        n_update = 0
 
             if with_progress:
-                progress_bar.set_postfix(
-                    train_loss=np.mean(train_losses_[-500:])
-                )
-                progress_bar.update(len(batch[0]))
+                if n_update:
+                    progress_bar.update(n_update)
+                progress_bar.close()
 
-        if with_progress:
-            progress_bar.close()
-
-        mean_train_loss = np.mean(train_losses_)
-        train_losses.append(mean_train_loss)
-        print_str += '{}Losses: train = {:.8f}' \
-                         .format(print_indent, mean_train_loss)
+            mean_train_loss = np.mean(train_losses_)
+            train_losses.append(mean_train_loss)
+            print_str += '{}Losses: train = {:.8f}' \
+                             .format(print_indent, mean_train_loss)
 
         test_losses_, test_golds, test_preds = [], [], []
 
@@ -495,37 +541,44 @@ def train(loaders, model, criterion, optimizer, scheduler,
         if test_loader:
             model.eval()
             for batch in test_loader:
-                batch = to_device(batch, device)
-                gold, gold_lens = batch[-1], batch[1]
-                [test_golds.extend(y_[:len_])
-                     for y_, len_ in zip(gold.cpu().numpy(), gold_lens)]
+                batch, gold, gold_lens = batch[:-1], batch[-1], batch[1]
+                if batch_to_device:
+                    batch = to_device(batch, device)
 
                 with torch.no_grad():
-                    pred = model(*batch[:-1])
+                    pred = model(*batch)
 
-                pred_values, pred_indices = pred.max(2)
+                pred_indices = pred.round().int() if is_bce else \
+                               pred.argmax(-1)
 
-                [test_preds.extend(y_[:len_])
-                     for y_, len_ in zip(pred_indices.cpu().numpy(), gold_lens)]
+                if len(pred.shape) == 2:
+                    test_golds.extend(gold.numpy())
+                    test_preds.extend(pred_indices.cpu().numpy())
+                else:
+                    [test_golds.extend(y_[:len_])
+                         for y_, len_ in zip(gold.numpy(),
+                                             gold_lens)]
+                    [test_preds.extend(y_[:len_])
+                         for y_, len_ in zip(pred_indices.cpu().numpy(),
+                                             gold_lens)]
 
-                batch_loss = []
-                for i in range(pred.size(0)):
-                    loss_ = criterion(pred[i], gold[i])
-                    batch_loss.append(loss_)
-
-                loss = torch.mean(torch.stack(batch_loss))
-                test_losses_.append(loss.item())
+                if criterion:
+                    if is_bce:
+                        gold = gold.float()
+                    loss = criterion(
+                        pred.flatten(end_dim=flatten_idx),
+                        to_device(gold, device).flatten(end_dim=-1)
+                    )
+                    test_losses_.append(loss.item())
 
             mean_test_loss = np.mean(test_losses_)
             test_losses.append(mean_test_loss)
 
             accuracy = accuracy_score(test_golds, test_preds)
-            precision = precision_score(test_golds, test_preds,
-                                        average='macro')
-            recall = recall_score(test_golds, test_preds,
-                                  average='macro')
-            f1 = f1_score(test_golds, test_preds,
-                          average='macro')
+            precision = \
+                precision_score(test_golds, test_preds, average='macro')
+            recall = recall_score(test_golds, test_preds, average='macro')
+            f1 = f1_score(test_golds, test_preds, average='macro')
 
             accuracies.append(accuracy)
             precisions.append(precision)
@@ -534,6 +587,8 @@ def train(loaders, model, criterion, optimizer, scheduler,
 
             score = -mean_test_loss if control_metric == 'loss' else \
                     accuracy if control_metric == 'accuracy' else \
+                    precision if control_metric == 'precision' else \
+                    recall if control_metric == 'recall' else \
                     f1 if control_metric == 'f1' else \
                     None
 
@@ -561,17 +616,31 @@ def train(loaders, model, criterion, optimizer, scheduler,
                 print_str += '\n{}BAD EPOCHS: {} ({})' \
                                  .format(log_prefix, bad_epochs_, sgn)
                 if bad_epochs_ >= bad_epochs and epoch >= min_epochs:
-                    print_str += ('\n{}Maximum bad epochs exceeded. Process '
-                                  'has been stopped. ').format(log_prefix) \
-                               + ('No models could surpass `best_score={}` given'
-                                      if best_epoch is None else
-                                  'Best score {} (on epoch {})') \
-                                     .format(best_score, best_epoch)
+                    print_str += \
+                        ('\n{}Maximum bad epochs exceeded. Process '
+                         'has been stopped. ').format(log_prefix) \
+                      + ('No models could surpass `best_score={}` given'
+                             if best_epoch is None else
+                         'Best score {} (on epoch {})').format(best_score,
+                                                               best_epoch)
                     break
 
+            if not train_loader:
+                break
+
+            if epoch == epochs:
+                print_str += \
+                    '\nMaximum epochs exceeded. ' \
+                    'Process has been stopped. ' \
+                  + ('No models could surpass `best_score={}` given'
+                         if best_epoch is None else
+                     'Best score {} (on epoch {})').format(best_score,
+                                                           best_epoch)
+
             if log_file:
                 print(print_str, file=log_file)
                 log_file.flush()
+                print_str = ''
 
             prev_score = score
 
@@ -579,10 +648,10 @@ def train(loaders, model, criterion, optimizer, scheduler,
             best_model_backup_method(model, score)
 
         if scheduler:
-            if score is None:
-                try:
-                    scheduler.step()
-                except TypeError:
+            try:
+                scheduler.step()
+            except TypeError:
+                if score is None:
                     scheduler.step(mean_train_loss)
                     warnings.warn(
                         ('The test score is None' if test_loader else
@@ -591,22 +660,23 @@ def train(loaders, model, criterion, optimizer, scheduler,
                       + '. The mean train loss have been used instead',
                         RuntimeWarning
                     )
-            else:
-                try:
+                else:
                     scheduler.step(score)
-                except TypeError:
-                    scheduler.step()
 
-    if log_file:
-        print(print_str, file=log_file)
+    if log_file and train_loader:
+        if print_str:
+            print(print_str, file=log_file)
+        print('Elapsed time: {}'
+                  .format(seconds_to_strtime(time.time() - start_time)),
+              file=log_file)
         log_file.flush()
 
-    return {'train_losses': train_losses,
-            'test_losses': test_losses,
-            'best_epoch': best_epoch,
+    return {'best_epoch': best_epoch,
             'best_score': best_score,
             'best_test_golds': best_test_golds,
             'best_test_preds': best_test_preds,
+            'train_losses': train_losses,
+            'test_losses': test_losses,
             'accuracies': accuracies,
             'precisions': precisions,
             'recalls': recalls,
